/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef POSTGRES_HPP_
#define POSTGRES_HPP_

#include <utility>
#include <vector>
#include <stdexcept>
#include <memory>
#include <optional>
#include <unordered_map>
#include <functional>
#include <string>
#include <cstring>
#include <cstdlib>
#include <unistd.h>
#include <sys/time.h>
#include <sys/select.h>

#if __has_include(<libpq-fe.h>)
#include <libpq-fe.h>
#elif __has_include(<pgsql/libpq-fe.h>)
#include <pgsql/libpq-fe.h>
#else
#include <postgresql/libpq-fe.h>
#endif

namespace sql {
using field_t = std::optional<std::string_view>;
struct error_t : std::runtime_error {
    inline explicit error_t(const char *str) : std::runtime_error(str) {}
    inline explicit error_t(PGconn **conn) : std::runtime_error(PQerrorMessage(*conn)) {
        PQfinish(*conn);
        *conn = nullptr;
    }
};

class results_t {
public:
    inline explicit results_t(PGresult *results) :
    results_(results) {
        if(!results)
            return;
        rows_ = PQntuples(results);
        cols_ = PQntuples(results);
    }

    inline ~results_t() {
        if(results_)
            PQclear(results_);
    }

    auto operator=(const results_t&) -> results_t& = delete;

    inline explicit operator bool() const {
        return results_ != nullptr;
    }

    inline auto operator!() const {
        return results_ == nullptr;
    }

    inline auto operator->() const {
        return results_;
    }

    inline auto operator[](const std::pair<int,int>& index) const -> field_t {
        auto [row, col] = index;
        if(row < 0 || row > rows_ || col < 0 || col > cols_)
            throw  std::out_of_range("invalid index");
        if(PQgetisnull(results_, row, col) != 0)
            return {};
        return std::string_view(PQgetvalue(results_, row, col), PQgetlength(results_, row, col));
    }

    inline auto operator[](int col) const -> field_t {
        auto row = 0;
        if(rows_ != 1)
            throw std::out_of_range("not single row");
        if(col < 0 || col > cols_)
            throw  std::out_of_range("invalid index");
        if(PQgetisnull(results_, row, col) != 0)
            return {};
        return std::string_view(PQgetvalue(results_, row, col), PQgetlength(results_, row, col));
    }

    inline auto headers() const {
        std::unordered_map<std::string_view, int> map;

        if(cols_ > 0) {
            for(int col = 0; col < cols_; ++col)
                map[PQfname(results_, col)] = col;
        }
        return map;
    }

    inline auto is_error() const {
        return !results_ || PQresultStatus(results_) != PGRES_COMMAND_OK;
    }

    inline void release() {
        if(results_) {
            PQclear(results_);
            results_ = nullptr;
        }
    }

    inline auto rows() const {
        return rows_;
    }

    inline auto cols() const {
        return cols_;
    }

protected:
    friend class connect_t;

    mutable PGresult *results_;
    int rows_{0}, cols_{0};

    inline results_t(results_t&& from) :
    results_(from.results_), rows_(from.rows_), cols_(from.cols_) {
        from.results_ = nullptr;
        from.rows_ = 0;
        from.cols_ = 0;
    }
};

class connect_t {
public:
    using notify_t = std::function<void(std::string_view,std::string_view)>;
    using yield_t = std::function<void(const results_t& results)>;

    inline connect_t() :
    notify_(nullptr) {}

    inline explicit connect_t(const std::string& uri, notify_t notify = nullptr) :
    notify_(std::move(notify)), conn_(PQconnectdb(uri.c_str())) {
        if(PQstatus(conn_) != CONNECTION_OK) {
            throw error_t(&conn_);
        }
        sock_ = PQsocket(conn_);
    }

    inline explicit operator PGconn *() { return conn_;}
    inline explicit operator bool() { return conn_ != nullptr;}
    inline auto operator!() { return conn_ == nullptr;}
    inline auto operator*() { return conn_;}

    inline connect_t(connect_t &&rhs) noexcept :
    notify_(std::move(rhs.notify_)), conn_(rhs.conn_), sock_(rhs.sock_) {
        rhs.notify_ = nullptr;
        rhs.conn_ = nullptr;
        rhs.sock_ = -1;
    }

    inline ~connect_t() {
        close();
    }

    connect_t(const connect_t& from) = delete;
    auto operator=(const connect_t& from) -> auto& = delete;

    inline auto operator=(connect_t &&rhs) noexcept -> connect_t & {
        std::swap(conn_, rhs.conn_);
        std::swap(sock_, rhs.sock_);
        std::swap(notify_, rhs.notify_);
        return *this;
    }

    inline auto connect(const std::string& uri, notify_t notify = nullptr) {
        close();
        conn_ = PQconnectdb(uri.c_str());
        if(!conn_ || PQstatus(conn_) != CONNECTION_OK) {
            if(conn_)
                PQfinish(conn_);
            conn_ = nullptr;
            return false;
        }
        sock_ = PQsocket(conn_);
        notify_ = notify;
        return true;
    }

    inline void close() {
        if(conn_)
            PQfinish(conn_);
        conn_ = nullptr;
        notify_ = nullptr;
    }

    inline auto version() const {
        return PQserverVersion(conn_);
    }

    inline auto user() const {
        return std::string(PQuser(conn_));
    }

    inline auto database() const {
        return std::string(PQdb(conn_));
    }

    inline auto socket() const {
        return sock_;
    }

    inline void reset() {
        PQreset(conn_);
        if(PQstatus(conn_) == CONNECTION_OK)
            sock_ = PQsocket(conn_);
        else
            throw error_t(&conn_);
    }

    inline void cancel() const {
        ::close(sock_);
    }

    inline auto escape(const std::string& request) {
        auto p = PQescapeLiteral(conn_, request.c_str(), request.size());
        if(!p)
            return std::string();
        std::string s(p);
        PQfreemem(p);
        return s;
    }

    inline void consumes() {
        PQconsumeInput(conn_);
        notifies();
    }

    inline void notifies() {
        PGnotify *notify = nullptr;
        while(nullptr != (notify = PQnotifies(conn_))) {
            if(notify_)
                notify_(std::string_view(notify->relname), std::string_view(notify->extra));
            PQfreemem(notify);
        }
    }

    inline void query(const std::string& sql, yield_t yield) {
        auto result = results_t(PQexec(conn_, sql.c_str()));
        yield(result);
        notifies();
    }

    inline auto query(const std::string& sql) {
        auto result = results_t(PQexec(conn_, sql.c_str()));
        notifies();
        return result;
    }

    static inline auto ping(const std::string& uri) {
        return PQping(uri.c_str()) == PQPING_OK;
    }

protected:
    notify_t notify_;
    PGconn *conn_{};
    int sock_{-1};
};
} // end namespace

/*!
 * Namespace for database services.
 * \namespace database
 */

/*!
 * Header only utility c++ library for postgresql server.
 * \file postgres.hpp
 */
#endif
