/*
 * Copyright (C) 2017-2019 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DEVICES_HPP_
#define DEVICES_HPP_

#include <filesystem.hpp>
#include <string>
#include <cstring>

namespace dev {
#ifdef OS_WINDOWS
inline auto is_disk(const std::string path) {
    if(path.length() == 2) {
        if((toupper(path[0]) < 'A') || (toupper(path[0]) > 'Z'))
            return false;

        return path[1] == ':';
    }

    auto devpath = path;
    if(path[0] != '/' && path[0] != '\\')
        devpath = "\\device\\" + path;
    if(devpath.substr(0, 8) == "/device/" || devpath.substr(0, 8) == "\\device\\")
        return fsys::is_block_file(devpath);
    else
        return false;
}

inline auto is_char(const std::string path) {
    if(path == "nul" || path == "nul:" || path == "con" || path == "con:" || path == "prn" || path == "prn:" || path == "aux" || path == "aux:")
        return true;

    if(path == "null")
        return true;

    if(path.substr(0, 3) == "com" || path.substr(0, 3) == "lpt") {
        size_t pos = 3;
        while(path[pos] >= '0' && path[pos] <= '9')
            ++pos;
        if(pos == path.length())
            return true;
        if(path[pos] == ':')
            return true;
    }

    auto devpath = path;
    if(path[0] != '/' && path[0] != '\\')
        devpath = "\\device\\" + path;
    if(devpath.substr(0, 8) == "/device/" || devpath.substr(0, 8) == "\\device\\")
        return fsys::is_character_file(devpath);

    return false;
}

inline auto device_pair(const std::string& path) -> std::pair<std::string, std::string> {
    if(path.length() == 2) {
        if((toupper(path[0]) < 'A') || (toupper(path[0]) > 'Z'))
            return {};

        if(path[1] == ':')
            return {path,{}};

        return {};
    }

    if(path == "null")
        return {"nul", {}};

    if(path == "nul" || path == "nul:" || path == "con" || path == "con:" || path == "prn" || path == "prn:" || path == "aux" || path == "aux:")
        return {path.substr(0, 3), {}};

    if(path.substr(0, 3) == "com" || path.substr(0, 3) == "lpt") {
        size_t pos = 3;
        while(path[pos] >= '0' && path[pos] <= '9')
            ++pos;
        if(pos == path.length())
            return {path, {}};
        if(path[pos] == ':')    // options...
            return {path.substr(0, pos - 1), path.substr(++pos)};
    }

    if(path[0] != '/' && path[0] != '\\')
        return {"\\device\\" + path, {}};
    if(path.substr(0, 8) == "/device/" || path.substr(0, 8) == "\\device\\")
        return {path, {}};
    return {};
}

#else
inline auto device_pair(const std::string& path) -> std::pair<std::string, std::string> {
    std::string devopt, devpath = path;

    auto pos = path.find(':');
    if(pos > 0) {
        devpath = path.substr(0, pos - 1);
        devopt = path.substr(++pos);
    }
    if(devpath.substr(0, 5) == "/dev/")
        return {devpath, devopt};

    return {"/dev/" + devpath, devopt};
}

inline auto is_char(const std::string path) {
    return fsys::is_character_file(device_pair(path).first);
}

inline auto is_disk(const std::string path) {
    return fsys::is_block_file(device_pair(path).first);
}
#endif

inline auto device_path(const fsys::path path) {
    return device_pair(std::string{path.u8string()}).first;
}

inline auto device_option(const std::string path) {
    return device_pair(path).second;
}
} // end namespace

#endif
