/*
 * Copyright (C) 2018-2019 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include <print.hpp>
#include <args.hpp>
#include <filesystem.hpp>
#include <fstream>

#include <openssl/rand.h>

#ifdef OS_WINDOWS
#include <windows.h>
#else
#include <sys/statvfs.h>
#endif

namespace {
args::option helpflag('h',  "--help", "display this list");
args::option althelp('?',   nullptr, nullptr);
args::option gigvalue('g',  nullptr, "output in gigabytes");
args::option kilovalue('k', nullptr, "output in kilobytes");
args::option megvalue('m',  nullptr, "output in megabytes");
args::option percent('p',   nullptr, "output in percentage");
args::option quiet('q',     "--quiet", "no output overwritting");
args::option randfill('r',  "--random", "fill space with hard random values");
args::option sizelimit('s', "--size", "space to zero or fill", "size");
args::option zerofill('z',  "--zero", "fill space with zero bytes");

int64_t space = 0, total = 0, iocount = 0, iosize = static_cast<int64_t>(64 * 1024);
std::string num, fillpath, path;
int exit_code = 0;

auto random_fill(uint8_t *buf, size_t size) {
    if(RAND_bytes(buf, static_cast<int>(size)))
        return size;
    return size_t(0);
}

void random_seed() {
    RAND_poll();
}

void cleanup()
{
#ifdef OS_WINDOWS
    if(fillpath.size() > 0)
        fsys::remove(fillpath);
#endif

    if(!iocount || !std::stoll(*sizelimit))
        return;

    if(is(quiet))
        return;

    for(unsigned count = 0; count < num.size(); ++count)
        print("{}", static_cast<char>(8));

    std::string suffix;
    if(is(kilovalue)) {
        suffix = "K";
        iocount >>= 10;
    }
    else if(is(megvalue)) {
        iocount >>= 20;
        suffix = "M";
    }
    else if(is(gigvalue)) {
        iocount >>= 30;
        suffix = "G";
    }
    print("{}{} overriden\n", iocount, suffix);
}

#ifdef OS_WINDOWS
BOOL CtrlHandler(DWORD event)
{
    switch(event) {
    case CTRL_C_EVENT:
    case CTRL_BREAK_EVENT:
        exit_code = 130;
        std::cout << "^C" << std::ends;
        sizelimit.set(0);
        return TRUE;
    default:
        exit_code = 143;
        sizelimit.set(0);
        return TRUE;
    }
}
#endif

} // end anon namespace

auto main(int argc, const char **argv) -> int
{
    try {
        path = ".";
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count > 1) {
            args::help({"freespace [options] [dir-path]"}, {
                "freespace " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Output free space of path or zero out free space"
            });
            return 0;
        }

        if(count > 0)
            path = args::list()[0];

        if(!fsys::is_directory(path))
            throw std::runtime_error("not a directory: " + path);

#ifdef OS_WINDOWS
        ULARGE_INTEGER v_space, v_total;
        if(!GetDiskFreeSpaceExA(path.c_str(), &v_space, nullptr, &v_total))
            throw std::runtime_error("cannot stat: " + path);
        space = static_cast<int64_t>(v_space.QuadPart);
        total = static_cast<int64_t>(v_total.QuadPart);
#else
        struct ::statvfs vfs{};
        if(statvfs(path.c_str(), &vfs) != 0)
            throw std::runtime_error("cannot stat: " + path);
        iosize = static_cast<int64_t>(vfs.f_bsize);
        space = static_cast<int64_t>(vfs.f_frsize) * static_cast<int64_t>(vfs.f_bavail);
        total = static_cast<int64_t>(vfs.f_frsize) * static_cast<int64_t>(vfs.f_blocks);
#endif
        if(!zerofill) {
            if(is(kilovalue))
                space >>= 10;
            else if(is(megvalue))
                space >>= 20;
            else if(is(gigvalue))
                space >>= 30;
            else if(is(percent))
                space = (int)(space / (total / 100));
            print("{}\n", space);
        }
        else {
            if(is(randfill))
                random_seed();
            char buf[iosize];
            fillpath = path + "/.zerofile";
            int64_t prior = -1, current = 0;

            memset(buf, 0, iosize);
            std::ofstream zs(fillpath, std::ofstream::binary);
            if(!zs.is_open())
                throw std::runtime_error("cannot write: " + path);
            // this only works in unix...
            fsys::remove(fillpath); // make sure file deletes at exit...
            atexit(cleanup);
#ifdef OS_WINDOWS
            SetConsoleCtrlHandler(CtrlHandler, TRUE);
#endif
            if(!is(sizelimit)) {
                sizelimit.set(strdup(std::to_string(space).c_str()));
                total = space;
            }
            else
                total = std::stoll(*sizelimit);
            while(zs.good() && iocount < std::stoll(*sizelimit)) {
                if(is(randfill))
                    random_fill(reinterpret_cast<uint8_t *>(buf), iosize);
                zs.write(buf, static_cast<std::streamsize>(iosize));
                if(zs.good())
                    zs.flush();
                if(zs.bad() || zs.fail())
                    break;
                iocount += iosize;
                current = iocount / (total / 100);
                if(current != prior && !is(quiet)) {
                    for(unsigned counter = 0; counter < num.size(); ++counter)
                        print("{}", static_cast<char>(8));

                    num = std::to_string(current) + "%";
                    print("{}", num);
                }
                prior = current;
            }
        }
    }
    catch(const bad_arg& e) {
        die(3, "Usage: freespace [options] [dir-path]\nerror: {}\n", e.what());
    }
    catch(const std::runtime_error& e) {
        die(2, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(1, "error: {}\n", e.what());
    }
#ifdef OS_WINDOWS
    if(fillpath.size() > 0)
        fsys::remove(fillpath);
#endif
    return exit_code;
}

