/*
 * Copyright (C) 2010-2014 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include <print.hpp>
#include <args.hpp>

#ifdef  OS_WINDOWS
#include <windows.h>
#include <conio.h>
#else
#include <sys/select.h>
#include <termios.h>
#endif

namespace {
args::option helpflag('h',  "--help", "display this list");
args::option althelp('?',   nullptr, nullptr);
args::option timeout('t',     "--timeout", "optional input timeout", "seconds");

#ifdef OS_WINDOWS
HANDLE input;
DWORD mode;
#else
struct termios orig, current;

void cleanup()
{
    tcsetattr(0, TCSANOW, &orig);
}
#endif

void keyflush()
{
#ifdef  OS_WINDOWS
    TCHAR ch;
    DWORD count;

    while(WaitForSingleObject(input, 20) == WAIT_OBJECT_0) {
        ReadConsole(input, &ch, 1, &count, nullptr);
    }
#else
    fd_set inp;
    struct timeval tv{};
    char ch{0};

    FD_ZERO(&inp);      // NOLINT
    FD_SET(0, &inp);    // NOLINT

    for(;;) {
        tv.tv_usec = 20000;
        tv.tv_sec = 0;
        if(select(1, &inp, nullptr, nullptr, &tv) <= 0)
            break;
        if(::read(0, &ch, 1) < 1)   // FlawFinder: ignore
            break;
    }
#endif
}
} // end namespace

auto main(int argc, const char **argv) -> int
{
    auto nl = false;
    auto excode = 2;

    try {
        auto count = args::parse(argc, argv);

        if(is(helpflag) || is(althelp) || count > 1) {
            args::help({"keywait [options] [\"prompt\"]"}, {
                "keywait " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Wait for input"
            });
            return 0;
        }

        if(count == 1) {
            nl = true;
            print("{}", args::list()[0]);
        }
        std::cout.flush();

#ifdef  OS_WINDOWS
        input = GetStdHandle(STD_INPUT_HANDLE);
        GetConsoleMode(input, &mode);
        SetConsoleMode(input, 0);

        keyflush();

        DWORD timer = INFINITE;

        if(is(timeout))
            timer = atoi(*timeout);

        if(WaitForSingleObject(input, timer) == WAIT_OBJECT_0) {
            keyflush();
            excode = 0;
        }
        else
            excode = 1;

        SetConsoleMode(input, mode);
#else
        tcgetattr(0, &orig);
        atexit(cleanup);
        tcgetattr(0, &current);
#ifdef  __sun
        current.c_iflag &= ~(IMAXBEL|IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
        current.c_oflag &= ~OPOST;
        current.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
        current.c_cflag &= ~(CSIZE|PARENB);
        current.c_cflag |= CS8;
#else
        cfmakeraw(&current);
#endif
        tcsetattr(0, TCSANOW, &current);
        fd_set inp;
        struct timeval tv = {0, 0};
        struct timeval *tvp = nullptr;

        FD_ZERO(&inp);      // NOLINT
        FD_SET(0, &inp);    // NOLINT


        if(*timeout) {
            tv.tv_sec = atoi(*timeout); // FlawFinder: ignore
            tvp = &tv;
        }

        int rtn = select(1, &inp, nullptr, nullptr, tvp);
        if(rtn > 0) {
            keyflush();
            excode = 0;
        }
        else {
            if(rtn == 0)
                excode = 1;
        }
        cleanup();
#endif
    }
    catch(const bad_arg& e) {
        die(3, "Usage: keywait [options] [\"prompt\"]\nerror: {}\n", e.what());
    }
    catch(const std::runtime_error& e) {
        die(2, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(1, "error: {}\n", e.what());
    }

    if(nl)
        print("\n");

    return excode;
}

