/*
 * Copyright (C) 2020 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include <print.hpp>
#include <digest.hpp>
#include <args.hpp>
#include <strings.hpp>
#include <filesystem.hpp>
#include <fstream>

using namespace crypto;

namespace {
inline constexpr char digits[17] = "0123456789abcdef";

args::option helpflag('h',  "--help", "display this list");
args::option althelp('?',   nullptr, nullptr);
args::option hashtype('d',  "--digest", "digest type to use", "type");
args::option sizeflag('s',  "--size", "output size of data");

auto to_hex(const uint8_t *data, size_t size) -> std::string
{
    std::ostringstream out;
    while(data && size--) {
        out << digits[*data >> 4] << digits[*data & 0x0f];
        ++data;
    }
    return out.str();
}

void output(digest_t& digest, std::istream& in, const std::string& name)
{
    char buffer[1024];
    auto total = 0UL;

    while(in && !in.eof()) {
        in.read(buffer, sizeof(buffer));    // FlawFinder: valid
        auto count = in.gcount();
        if(count < 1)
            break;
        total += count;
        digest.update(buffer, static_cast<size_t>(count));
    }
    digest.finish();
    print("{}", to_hex(digest.data(), digest.size()));
    if(is(sizeflag))
        print(" {}", total);
    print(" {}\n", name);
    digest.reinit();
}

} // end anon namespace

auto main(int argc, const char **argv) -> int
{
    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp)) {
            args::help({"mdsum [options] args.."}, {
                "mdsum " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Hash digest of specified files"
            });
            return 0;
        }

        if(!is(hashtype))
            hashtype.set("sha256");

        const EVP_MD *md = nullptr;
        auto hashid = lower_case(std::string(*hashtype));
        if(hashid == "sha256" || hashid == "sha2" || hashid == "256")
            md = EVP_sha256();
        else if(hashid == "sha224" || hashid == "224")
            md = EVP_sha224();
        else if(hashid == "sha384" || hashid == "384")
            md = EVP_sha384();
        else if(hashid == "sha512" || hashid == "512")
            md = EVP_sha512();
        else if(hashid == "md5" || hashid == "128")
            md = EVP_md5();
        else if(hashid == "sha" || hashid == "sha1")
            md = EVP_sha1();
        else
            md = EVP_get_digestbyname(hashid.c_str());

        if(!md)
            throw std::out_of_range("invalid hash type: " + hashid);

        digest_t digest(md);

        if(count < 1) {
            output(digest, std::cin, "-");
            return 0;
        }

        for(auto const& arg : args::list()) {
            auto path = fsys::path(arg);
            if(fsys::is_directory(path)) {
                auto files = fsys::directory_iterator(path);
                for(const auto& entry : files) {
                    auto name = std::string{entry.path().filename().u8string()};
                    auto stat = fsys::status(entry);
                    if(fsys::is_directory(stat))
                        continue;
                    if(name[0] == '.')
                        continue;
                    std::ifstream file(entry.path(), std::ios::binary);
                    if(file.is_open())
                        output(digest, file, name);
                }
            }
            else {
                std::ifstream file(path, std::ios::binary);
                if(!file.is_open())
                    throw std::runtime_error(std::string{path.u8string()} + ": not accessible");
                output(digest, file, path.filename().u8string());
            }
        }
    }
    catch(const bad_arg& e) {
        die(3, "Usage: mdsum [options] args..\nerror: {}\n", e.what());
    }
    catch(const std::runtime_error& e) {
        die(2, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(1, "error: {}\n", e.what());
    }
    return 0;
}

