/*
 * Copyright (C) 2018-2019 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include <args.hpp>
#include <filesystem.hpp>
#include <print.hpp>
#include <list>

#include <algorithm>

namespace {
args::option helpflag('h',  "--help", "display this list");
args::option althelp('?',   nullptr, nullptr);
args::option basename('b',  "--basename", "reduce paths to base filename");
args::option delim('d',     "--delim", "set deliminter between arguments", "string");
args::option directory('D', "--directory", "expand directory into file arguments");
args::option lines('l',     "--lines", "list arguments on separate lines");
args::option quoting('q',   "--quote",  "set quote for each argument", "pair");
args::option recursive('R', "--recursive", "recursive directory scan");
args::option follow('F',    "--follow", "follow symlinks");
args::option rflag('r',     "--reverse", "reverse order of arguments");
args::option sortflag('s',  "--sort", "sort directory");
std::string prefix, suffix;
bool middle = false;

void output(const fsys::path& arg)
{
    std::string name{arg.u8string()};
    if(is(basename))
        name = std::string{arg.filename().u8string()};
    if(is(lines))
        print("{}{}{}\n", prefix, name, suffix);
    else if(middle && delim)
        print("{}{}{}{}", *delim, prefix, name, suffix);
    else if(middle)
        print(" {}{}{}", prefix, name, suffix);
    else
        print("{}{}{}", prefix, name, suffix);
    middle = true;
}

void scan(const fsys::path& path, bool top = true)
{
    auto dir = fsys::directory_iterator(path);
    auto count = 0;

    std::list<std::string> list;
    for(auto& p : fsys::directory_iterator(path))
        list.emplace_back(std::string{p.path().filename().u8string()});

    if(is(sortflag))
        list.sort();

    if(is(rflag))
        std::reverse(list.begin(), list.end());

    for(const auto & filename : list) {
        auto status = fsys::status(filename);
        if(filename[0] == '.')
            continue;
        ++count;
        auto subdir = fsys::path(path) / filename;
        output(subdir);
        if(fsys::is_directory(status)) {
            if(is(follow) || is(recursive)) {
                if(!fsys::is_symlink(subdir) || is(follow))
                    scan(subdir, false);
            }
        }
    }
    if(top && !count)
        output(path);
}
} // end anon namespace

auto main(int argc, const char **argv) -> int
{
    try {
        auto count = args::parse(argc, argv);

        if(is(helpflag) || is(althelp)) {
            args::help({"paths [options] args.."}, {
                "paths " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Echo pathname command line arguments"
            });
            return 0;
        }

        if(count < 1 && !is(directory) && !is(recursive))
            return 0;

        auto quote = *quoting;
        if(quote && quote[0]) {
            if(!quote[1])
                prefix = suffix = std::string(1, quote[0]);
            else {
                prefix = std::string(1, quote[0]);
                suffix = std::string(1, quote[1]);
            }
        }

        std::list<std::string> list;
        for(const auto& arg : args::list())
            list.emplace_back(arg);
        if(list.empty() && (is(directory) || is(recursive)))
            list.emplace_back(".");

        if(is(sortflag))
            list.sort();

        if(is(rflag))
            std::reverse(list.begin(), list.end());

        for(auto const& arg : list) {
            if(fsys::is_directory(arg) && (is(directory) || is(recursive) || is(follow)))
                scan(arg);
            else
                output(arg);
        }

        if(!is(lines))
            print("\n");
    }
    catch(const bad_arg& e) {
        die(3, "Usage: paths [options] args..\nerror: {}\n", e.what());
    }
    catch(const std::runtime_error& e) {
        die(2, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(1, "error: {}\n", e.what());
    }
    return 0;
}

