/*
 * Copyright (C) 2018-2019 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "devices.hpp"
#include <args.hpp>
#include <print.hpp>
#include <fstream>
#include <cstdlib>
#include <csignal>

#ifdef OS_WINDOWS
#include <windows.h>
#include <process.h>
#include <io.h>
using fd_t = HANDLE;
inline const auto FD_INVALID = reinterpret_cast<HANDLE>(INVALID_HANDLE_VALUE);
#else
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
using fd_t = int;
inline constexpr auto FD_INVALID = (-1);
#endif

using namespace std::string_literals;

namespace {
args::option helpflag('h',  "--help", "display this list");
args::option althelp('?',   nullptr, nullptr);
args::option logpath('l',   "--log", "path for logging stderr", "path");

#ifdef OS_WINDOWS
fd_t input_from(const std::string& path)
{
    SECURITY_ATTRIBUTES sattr;

    sattr.nLength = sizeof(SECURITY_ATTRIBUTES);
    sattr.bInheritHandle = TRUE;
    sattr.lpSecurityDescriptor = nullptr;

    return CreateFile(path.c_str(), GENERIC_READ, FILE_SHARE_READ, &sattr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, nullptr);
}

fd_t output_to(const std::string& path)
{
    SECURITY_ATTRIBUTES sattr;

    sattr.nLength = sizeof(SECURITY_ATTRIBUTES);
    sattr.bInheritHandle = TRUE;
    sattr.lpSecurityDescriptor = nullptr;

    return CreateFile(path.c_str(), GENERIC_WRITE, 0, &sattr, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, nullptr);
}

fd_t append_to(const std::string& path)
{
    SECURITY_ATTRIBUTES sattr;

    sattr.nLength = sizeof(SECURITY_ATTRIBUTES);
    sattr.bInheritHandle = TRUE;
    sattr.lpSecurityDescriptor = nullptr;

    fd_t fd = CreateFile(path.c_str(), GENERIC_WRITE, 0, &sattr, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, nullptr);
    if(fd != FD_INVALID)
        SetFilePointer(fd, 0, nullptr, FILE_END);

    return fd;
}

auto detach_exec(const char **args, const fd_t *stdio)
{
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    char filename[128];
    int pos;

    memset(&si, 0, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);

    //pathfinder(args[0], filename, sizeof(filename));
    snprintf(filename, sizeof(filename), "%s", args[0]);

    auto argc = 0;
    auto argv = new char[32768];
    argv[0] = 0;
    char *cp = argv;
    while(args && args[argc]) {
        if(argc)
            *(cp++) = ' ';
        snprintf(cp, 32767 - static_cast<unsigned>(cp - argv), "%s", args[argc++]);
        cp += strlen(cp);   // FlawFinder: ignore
    }

    for(pos = 0; pos < 3; ++pos) {
        fd_t fd = stdio[pos];
        switch(pos) {
        case 0:
            if(fd == FD_INVALID)
                fd = input_from("nul");
            si.hStdInput = stdio[pos];
            break;
        case 1:
            if(fd == FD_INVALID)
                fd = output_to("nul");
            si.hStdOutput = stdio[pos];
            break;
        case 2:
            if(fd == FD_INVALID)
                fd = output_to("nul");
            si.hStdError = stdio[pos];
            break;
        }
    }
    si.dwFlags = STARTF_USESTDHANDLES;
    if(!CreateProcess((CHAR *)filename, (CHAR *)args, nullptr, nullptr, TRUE, DETACHED_PROCESS | CREATE_NEW_PROCESS_GROUP, nullptr, nullptr, &si, &pi))
        throw std::runtime_error("cannot create process");
    CloseHandle(pi.hProcess);
    return 0;
}

#else
auto detach_exec(const char **args, const fd_t *stdio)
{
    auto pid = fork();
    if(pid < 0)
        throw std::runtime_error("cannot fork");
    if(pid > 0)
        return 0;

    ::signal(SIGQUIT, SIG_DFL);
    ::signal(SIGINT, SIG_DFL);
    ::signal(SIGCHLD, SIG_DFL);
    ::signal(SIGPIPE, SIG_DFL);
    ::signal(SIGHUP, SIG_DFL);
    ::signal(SIGABRT, SIG_DFL);
    ::signal(SIGUSR1, SIG_DFL);

#ifdef  SIGTTOU
    ::signal(SIGTTOU, SIG_IGN); // NOLINT
#endif

#ifdef  SIGTTIN
    ::signal(SIGTTIN, SIG_IGN); // NOLINT
#endif

#ifdef  SIGTSTP
    ::signal(SIGTSTP, SIG_IGN); // NOLINT
#endif

    fd_t fd{-1};
    for(fd = 0; fd < 3; ++fd) {
        if(stdio[fd] != FD_INVALID)
            ::dup2(stdio[fd], fd);
        else
            ::close(fd);
    }

    for(fd = 3; fd < static_cast<int>(sizeof(fd_set) * 8); ++fd)
        ::close(fd);

#if defined(SIGTSTP) && defined(TIOCNOTTY)
    if(setpgid(0, getpid()) == -1)
        ::exit(-1);

    // FlawFinder: ignore
    if((fd = open("/dev/tty", O_RDWR)) >= 0) {
        ::ioctl(fd, TIOCNOTTY, nullptr);
        ::close(fd);
    }
#else

#ifdef  HAVE_SETPGRP
    if(setpgrp() == -1)
        ::exit(-1);
#else
    if(setpgid(0, getpid()) == -1)
        ::exit(-1);
#endif
    if(getppid() != 1) {
        if((pid = fork()) < 0)
            ::exit(-1);
        else if(pid > 0)
            ::exit(0);
    }
#endif
    for(fd = 0; fd < 3; ++fd) {
        if(stdio[fd] != FD_INVALID)
            continue;
        // FlawFinder: ignore
        fd_t tmp = ::open("/dev/null", O_RDWR);
        if(tmp != fd) {
            ::dup2(tmp, fd);
            ::close(tmp);
        }
    }
    // FlawFinder: valid exec
    ::execvp(args[0], (char *const *)args); // NOLINT
    ::exit(-1);
}

auto input_from(const std::string& path) -> fd_t
{
    // FlawFinder: ignore
    return ::open(path.c_str(), O_RDONLY);
}

auto output_to(const std::string& path)-> fd_t
{
    // FlawFinder: ignore
    return ::open(path.c_str(), O_WRONLY | O_CREAT | O_TRUNC, 0640);
}

auto append_to(const std::string& path) -> fd_t
{
    // FlawFinder: ignore
    return ::open(path.c_str(), O_WRONLY | O_CREAT | O_APPEND, 0640);
}
#endif

} // namespace

auto main(int argc, const char **argv) -> int
{
    fd_t stdio[3];
#ifdef OS_WINDOWS
    stdio[0] = GetStdHandle(STD_INPUT_HANDLE);
    stdio[1] = GetStdHandle(STD_OUTPUT_HANDLE);
    stdio[2] = GetStdHandle(STD_ERROR_HANDLE);
#else
    stdio[0] = 0;
    stdio[1] = 1;
    stdio[2] = 2;
#endif

    try {
        auto opts = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || opts < 2) {
            args::help({"pattach [options] ttypath|none command ..."}, {
                "pattach " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Attach process to a tty and daemonize"
            });
            return 0;
        }

        if(is(logpath)) {
            stdio[2] = append_to(*logpath);
            if(stdio[2] == FD_INVALID)
                throw std::runtime_error("not a valid log path: "s + *logpath);
        }

        auto cli = args::list();
        if(cli[0] == "none" || cli[0] == "no") {
            stdio[0] = stdio[1] = FD_INVALID;
            if(!is(logpath))
                stdio[2] = FD_INVALID;
        }
        else {
            auto devpath = dev::device_path(cli[0]);
            if(devpath.empty())
                throw std::runtime_error("non-valid device: " + devpath);

            if(!dev::is_char(devpath))
                throw std::runtime_error("non-character device: "s + devpath);
            stdio[0] = input_from(devpath);
            stdio[1] = output_to(devpath);
        }

        auto args = new const char*[opts]; // NOLINT
        auto count = 1U;
        while(count < static_cast<unsigned>(cli.size())) {
            args[count - 1] = cli[count].data();    // cli points to argv...
            ++count;
        }
        args[count - 1] = nullptr;
        throw detach_exec(args, stdio);
    }
    catch(const bad_arg& e) {
        die(3, "Usage: pattach [options] ttypath|none command ...\nerror: {}\n", e.what());
    }
    catch(const std::runtime_error& e) {
        die(2, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(1, "error: {}\n", e.what());
    }
    catch(const int& exiting) {
        ::exit(exiting);
    }
    return 0;
}

